/*****************************************************************************/
/* DLM   RSL                                                        */   
/* ப  openssl  RSL                                             */
/*  0.0.0d (11.03.2023)                                                */
/* tema@mail.ru                                                              */
/*****************************************************************************/

#include <windows.h>
#include <rsl/dlmintf.h>
#include <openssl/evp.h>
#include <openssl/engine.h>
#include <openssl/crypto.h>
#include <openssl/conf.h>
#include <openssl/err.h>
#include <openssl/bio.h>
#include <openssl/pem.h>
#include <openssl/x509.h>
#include <openssl/x509v3.h>
#include <../bn/bn_lcl.h>
#include <stdio.h>
#include <io.h>

#define FILE_MAX_CHUNK_SIZE 524288
#define X509_NAME_MAXLEN 64

static void opselHash(void) {

    const char * sGostPrefix="md_gost";
    EVP_MD_CTX *mdctx;
    const EVP_MD *md;
    unsigned char md_value[EVP_MAX_MD_SIZE];
    long lErrLine;

    VALUE *vStr;
    VALUE *vDigestName;

        char* s;
        char* sDigestName;

        if (GetParm (1,&vStr) && vStr->v_type  == V_STRING) { 
            s=vStr->value.string;
            }
        else {
            if (vStr->v_type  == V_UNDEF) {
                void * pv=NULL;
                ReturnVal(V_UNDEF,&pv);
                }
            else
                RslError("ࠬ 2   ப");
            return;
            }

        if (GetParm (0,&vDigestName) && vDigestName->v_type  == V_STRING) { 
            sDigestName=vDigestName->value.string;
            }
        else {
            sDigestName="md5";
            }

    ENGINE *eGost = NULL;

    if(!OPENSSL_init_crypto(OPENSSL_INIT_ENGINE_ALL_BUILTIN | OPENSSL_INIT_ENGINE_DYNAMIC, NULL))
        RslError("訡  樠樨 ਯ⮡⥪");
    
    if (strlen(sDigestName)>=strlen(sGostPrefix) && 0==memcmp(sGostPrefix,sDigestName,strlen(sGostPrefix)))
        if((eGost = ENGINE_by_id("gost"))==NULL) 
            RslError(" 㦥    openssl 室  -㭪樨 %s",sDigestName);
        
    md = EVP_get_digestbyname(sDigestName);
    if (!md)
        RslError("-㭪 %s  ", sDigestName);

    mdctx = EVP_MD_CTX_new();
    if (!EVP_DigestInit_ex(mdctx, md, eGost))
        RslError("訡 樠樨   ஢");
    EVP_DigestUpdate(mdctx, s, strlen(s));
    unsigned int md_len;
    EVP_DigestFinal_ex(mdctx, md_value, &md_len);
    EVP_MD_CTX_free(mdctx);

    char * s_out =(char *)malloc(sizeof(char)*md_len*2+sizeof(char));
    if (!s_out)
        RslError("訡 ");

    for (int i = 0; i < md_len; i++)
        sprintf(s_out+i*2, "%02x", md_value[i]);
    s_out[md_len*2]='\0';

    ReturnVal(V_STRING,s_out);

    free(s_out);

}

static void opselHashFile(void) {

    const char * sGostPrefix="md_gost";
    EVP_MD_CTX *mdctx;
    const EVP_MD *md;
    unsigned char md_value[EVP_MAX_MD_SIZE];
    long lErrLine;

    VALUE *vFilePath; // vFilePath->value.string
    VALUE *vDigestName;

    char* sDigestName;

    if (!GetParm (1,&vFilePath) && vFilePath->v_type  == V_STRING)
        RslError("ࠬ 2   ப");
    char * sPath=(char *)malloc(sizeof(char)*strlen(vFilePath->value.string)+sizeof(char));
    OemToCharBuff(vFilePath->value.string, sPath, strlen(vFilePath->value.string));
    sPath[strlen(vFilePath->value.string)]='\0';

    if (GetParm (0,&vDigestName) && vDigestName->v_type  == V_STRING) { 
        sDigestName=vDigestName->value.string;
        }
    else {
        sDigestName="md5";
        }

    size_t result;

    FILE * pFile;
    if((pFile = fopen(sPath, "rb"))==NULL)
        RslError("訡  䠩 %s",sPath);

    fseek(pFile, 0, SEEK_END);
    long lFileSize = ftell(pFile);
    rewind (pFile);

    long lBuffSize = lFileSize;
    if (lBuffSize>FILE_MAX_CHUNK_SIZE)
        lBuffSize=FILE_MAX_CHUNK_SIZE;

    char * buffer;
    if((buffer = (char*) malloc (sizeof(char)*lBuffSize))==NULL)
        RslError("訡  %l",lBuffSize);

    ENGINE *eGost = NULL;

    if(!OPENSSL_init_crypto(OPENSSL_INIT_ENGINE_ALL_BUILTIN | OPENSSL_INIT_ENGINE_DYNAMIC, NULL))
        RslError("訡  樠樨 ਯ⮡⥪");
    
    if (strlen(sDigestName)>=strlen(sGostPrefix) && 0==memcmp(sGostPrefix,sDigestName,strlen(sGostPrefix)))
        if((eGost = ENGINE_by_id("gost"))==NULL) 
            RslError(" 㦥    openssl 室  -㭪樨 %s",sDigestName);
        
    md = EVP_get_digestbyname(sDigestName);
    if (!md)
        RslError("-㭪 %s  ", sDigestName);

    mdctx = EVP_MD_CTX_new();
    if(!EVP_DigestInit_ex(mdctx, md, eGost))
        RslError("訡 樠樨   ஢");

    while (size_t chunkSize=fread(buffer, sizeof(char), lBuffSize, pFile)) {
        EVP_DigestUpdate(mdctx, buffer, chunkSize);
        }

    unsigned int md_len;
    EVP_DigestFinal_ex(mdctx, md_value, &md_len);
    EVP_MD_CTX_free(mdctx);

    char * s_out =(char *)malloc(sizeof(char)*md_len*2+sizeof(char));
    if (!s_out)
        RslError("訡 ");

    for (int i = 0; i < md_len; i++)
        sprintf(s_out+i*2, "%02x", md_value[i]);
    s_out[md_len*2]='\0';

    fclose (pFile);
    free (buffer);

    ReturnVal(V_STRING,s_out);

}

void convert_ASN1TIME_ymd(ASN1_TIME* t, char* buf) {
    const char* str = (const char*) t->data;
    size_t i = 0;

    if (t->type == V_ASN1_UTCTIME) { // two digit year
        //int year = (str[i++] - '0') * 10 + (str[++i] - '0');
        //if (year < 70)
        //   year += 100;
        strcpy(buf,"20");
        strcat(buf,str);

        }
    else if (t->type == V_ASN1_GENERALIZEDTIME) { // four digit year
        strcpy(buf,str);
        }
}

//20190918140629
//YYYYMMDDhhmmss
//01234567890123
//          1111
void convert_ASN1TIME_to_RSDT(ASN1_TIME* t, RslDtTm * r) {
    const char* s = (const char*) t->data;
    size_t i = 0;

    if (t->type == V_ASN1_UTCTIME) { // two digit year
        r->date.day   =(s[4]-'0')*10+(s[5]-'0');
        r->date.mon   =(s[2]-'0')*10+(s[3]-'0');
        r->date.year  =(s[0]-'0')*10+(s[1]-'0')+2000;
        r->time.hundr =0;
        r->time.sec   =(s[10]-'0')*10+(s[11]-'0');
        r->time.min   =(s[8]-'0')*10+(s[9]-'0');
        r->time.hour  =(s[6]-'0')*10+(s[7]-'0');
        }
    else if (t->type == V_ASN1_GENERALIZEDTIME) { // four digit year
        r->date.day   =(s[6]-'0')*10+(s[7]-'0');
        r->date.mon   =(s[4]-'0')*10+(s[5]-'0');
        r->date.year  =(s[0]-'0')*1000+(s[1]-'0')*100+(s[2]-'0')*10+(s[3]-'0');
        r->time.hundr =0;
        r->time.sec   =(s[12]-'0')*10+(s[13]-'0');
        r->time.min   =(s[10]-'0')*10+(s[11]-'0');
        r->time.hour  =(s[8]-'0')*10+(s[9]-'0');
        }
}


#define DATE_LEN 128

void convert_ASN1TIME(ASN1_TIME *t, char* buf, size_t len) {
    int rc;
    BIO *b = BIO_new(BIO_s_mem());
    rc = ASN1_TIME_print(b, t);
    if (rc <= 0) {
        BIO_free(b);
        RslError("訡 ८ࠧ ASN1_TIME_print");
    }
    rc = BIO_gets(b, buf, len);
    if (rc <= 0) {
        BIO_free(b);
        RslError("訡  BIO_gets");
    }
    BIO_free(b);
}


char * BmpTo866(const wchar_t * s, size_t nBytes) {
    char *  sRet;
    char * p= (char *)s;
    char * buff = (char *)iNewMem(nBytes+2);

    for(int i=0;i<nBytes;i++,i++) {
        buff[i]=p[i+1];
        buff[i+1]=p[i];
    }

    sRet    = (char *)malloc(nBytes/2+1);
    sRet[nBytes/2]='\0';

    WideCharToMultiByte(866, 0, (const wchar_t *)buff, nBytes/2, sRet, nBytes/2, NULL, false);
    
    iDoneMem(buff);

    return sRet;
}

// must free
char * UTF8To866(const char * s) {
    BSTR    bstrWide;
    char *  sRet;
    int     nLength;

    nLength = MultiByteToWideChar(CP_UTF8, 0, s, strlen(s) + 1, NULL, NULL);
    bstrWide = SysAllocStringLen(NULL, nLength);

    MultiByteToWideChar(CP_UTF8, 0, s, strlen(s) + 1, bstrWide, nLength);

    nLength = WideCharToMultiByte(CP_ACP, 0, bstrWide, -1, NULL, 0, NULL, NULL);
    sRet    = (char *)malloc(nLength+1);
    sRet[nLength]='\0';

    WideCharToMultiByte(CP_ACP, 0, bstrWide, -1, sRet, nLength, NULL, NULL);
    SysFreeString(bstrWide);

    CharToOemBuff(sRet, sRet, nLength);

    return sRet;
}

void exportX509date(ASN1_TIME * xdt, const char * key, VALUE * vArrayKeys, VALUE * vArrayValues, int * pIndexElement) {
    RslDtTm rsdt;

    // ८ࠧ㥬 -६  ଠ ASN1  ଠ RS
    convert_ASN1TIME_to_RSDT(xdt, &rsdt);

    // ࠭塞  ᨢ  
    VALUE * pArrayElement=PushValue(NULL);
    ValueSet (pArrayElement,V_STRING,(void *)key);
    RslTArrayPut((TGenObject *)vArrayKeys->value.obj, *pIndexElement, pArrayElement);
    PopValue();

    // ࠭塞  ᨢ 祭 
    pArrayElement=PushValue(NULL);
    ValueSet (pArrayElement,V_DTTM,(void *)&rsdt);
    RslTArrayPut((TGenObject *)vArrayValues->value.obj, *pIndexElement, pArrayElement);
    PopValue();

    (*pIndexElement)++;
}


void exportX509kv(const char * key, const char * value, VALUE * vArrayKeys, VALUE * vArrayValues, int * pIndexElement) {
    RslDtTm rsdt;

    // ࠭塞  ᨢ  
    VALUE * pArrayElement=PushValue(NULL);
    ValueSet (pArrayElement,V_STRING,(void *)key);
    RslTArrayPut((TGenObject *)vArrayKeys->value.obj, *pIndexElement, pArrayElement);
    PopValue();

    // ࠭塞  ᨢ 祭 
    pArrayElement=PushValue(NULL);
    ValueSet (pArrayElement,V_STRING,(void *)value);
    RslTArrayPut((TGenObject *)vArrayValues->value.obj, *pIndexElement, pArrayElement);
    PopValue();

    (*pIndexElement)++;
}

void exportX509kv_int(const char * key, int intvalue, VALUE * vArrayKeys, VALUE * vArrayValues, int * pIndexElement) {
    RslDtTm rsdt;

    // ࠭塞  ᨢ  
    VALUE * pArrayElement=PushValue(NULL);
    ValueSet (pArrayElement,V_STRING,(void *)key);
    RslTArrayPut((TGenObject *)vArrayKeys->value.obj, *pIndexElement, pArrayElement);
    PopValue();

    // ࠭塞  ᨢ 祭 
    pArrayElement=PushValue(NULL);
    ValueSet (pArrayElement,V_INTEGER,&intvalue);
    RslTArrayPut((TGenObject *)vArrayValues->value.obj, *pIndexElement, pArrayElement);
    PopValue();

    (*pIndexElement)++;
}

void exportX509kv_bin(const char * key, int len, const unsigned char* bin, VALUE * vArrayKeys, VALUE * vArrayValues, int * pIndexElement) {
    
    char * buff = (char * )malloc (len*2+1);
    int i;

    for (i = 0; i < len; i++) 
        sprintf(buff + i * 2, "%02X", bin[i]);
    
    exportX509kv(key, buff, vArrayKeys, vArrayValues, pIndexElement);
    free(buff);
}

void exportX509Name(X509_NAME * xname, const char * xnamePrefix, VALUE * vArrayKeys, VALUE * vArrayValues, int * pIndexElement) {

    //char * sOgrnIp_oid="1.2.643.100.5";
    //char * sOgrnIp_str="OGRNIP";

    OBJ_create("1.2.643.100.5", "OGRNIP", "OGRNIP");

    for (int i = 0; i < X509_NAME_entry_count(xname); i++) {
        X509_NAME_ENTRY *e = X509_NAME_get_entry(xname, i);
        ASN1_OBJECT *o = X509_NAME_ENTRY_get_object(e);
        char namebuf[X509_NAME_MAXLEN];
        int buflen = OBJ_obj2txt(namebuf, sizeof(namebuf), o, 0);

        //if (strcmp(sOgrnIp_oid,namebuf)==0)
        //    strncpy (namebuf, sOgrnIp_str, X509_NAME_MAXLEN);

        ASN1_STRING *d = X509_NAME_ENTRY_get_data(e);
        const unsigned char *str = ASN1_STRING_get0_data(d);
        const char * sValue;
        char * str866=NULL;
        switch (d->type) {
            case V_ASN1_UTF8STRING:
                str866=UTF8To866((const char *)str);
                sValue=str866;
                break;
            case V_ASN1_BMPSTRING:
                str866=BmpTo866((const wchar_t *)d->data,d->length); // d->data
                sValue=str866;
                break;
            default:
                sValue=(const char *)str;
        }

        // ࠭塞  ᨢ  
        VALUE * pArrayElement=PushValue(NULL);
        char * key = (char *)iNewMem(strlen(namebuf)+strlen(xnamePrefix)+2);
        strcpy(key,xnamePrefix);
        strcat(key,namebuf);
        ValueSet (pArrayElement,V_STRING,(void *)key);
        RslTArrayPut((TGenObject *)vArrayKeys->value.obj, *pIndexElement, pArrayElement);
        PopValue();

        // ࠭塞  ᨢ 祭 
        pArrayElement=PushValue(NULL);
        ValueSet(pArrayElement,V_STRING,(void *)sValue);
        RslTArrayPut((TGenObject *)vArrayValues->value.obj, *pIndexElement, pArrayElement);
        PopValue();

        (*pIndexElement)++;

        iDoneMem(key);
        if (str866)
            free(str866);
    }
}

void exportX509kv_ASN1Obj(const char * key, ASN1_OBJECT *o, VALUE * vArrayKeys, VALUE * vArrayValues, int * pIndexElement) {

    OBJ_create("1.2.643.100.5", "OGRNIP", "OGRNIP");

    char namebuf[X509_NAME_MAXLEN];
    int buflen = OBJ_obj2txt(namebuf, sizeof(namebuf), o, 0);

    // ࠭塞  ᨢ  
    VALUE * pArrayElement=PushValue(NULL);
    ValueSet (pArrayElement,V_STRING,(void *)key);
    RslTArrayPut((TGenObject *)vArrayKeys->value.obj, *pIndexElement, pArrayElement);
    PopValue();

    // ࠭塞  ᨢ 祭 
    pArrayElement=PushValue(NULL);
    ValueSet(pArrayElement,V_STRING,(void *)namebuf);
    RslTArrayPut((TGenObject *)vArrayValues->value.obj, *pIndexElement, pArrayElement);
    PopValue();

    (*pIndexElement)++;
}


void fingerprint(X509 * x509, VALUE * vArrayKeys, VALUE * vArrayValues, int * pIndexElement) {
    int j;
    unsigned int md_len;
    unsigned char md_value[EVP_MAX_MD_SIZE];

    if (!X509_digest(x509, EVP_sha1(), md_value, &md_len)) {
        return;
    }

    char * s_out = (char *)malloc(sizeof(char)*md_len*2+sizeof(char));
    if (!s_out)
        RslError("訡 ");

    for (int i = 0; i < md_len; i++)
        sprintf(s_out+i*2, "%02x", md_value[i]);
    s_out[md_len*2]='\0';

    exportX509kv("fingerprint", s_out, vArrayKeys, vArrayValues, pIndexElement);

    free(s_out);
}


char * rsGetFilePathParam(int iParam) {
    VALUE *vFilePath;
    if (!GetParm (iParam,&vFilePath) && vFilePath->v_type  == V_STRING)
        RslError("ࠬ %i   ப",(iParam+1));
    char * sPath=(char *)malloc(sizeof(char)*strlen(vFilePath->value.string)+sizeof(char));
    OemToCharBuff(vFilePath->value.string, sPath, strlen(vFilePath->value.string));
    sPath[strlen(vFilePath->value.string)]='\0';
    return sPath;
}



X509 * rsGetX509(int iParam) {
    VALUE *vCertPath;
    char *sCertPath;

    if (GetParm (iParam, &vCertPath) && vCertPath->v_type  == V_STRING) { 
        sCertPath=(char *)iNewMem(sizeof(char)*strlen(vCertPath->value.string)+sizeof(char));
        OemToCharBuff(vCertPath->value.string, sCertPath, strlen(vCertPath->value.string));
        sCertPath[strlen(vCertPath->value.string)]='\0';
        }
    else
        RslError("ࠬ %i   ப c  䠩 䨪",(iParam+1));

    char * certBuff = NULL;
    long certLen;
    FILE *fCert = fopen(sCertPath, "r");

    if (!fCert) 
        RslError("訡  ⨨ 䠩 䨪 %s",sCertPath);

    iDoneMem(sCertPath);

    fseek (fCert, 0, SEEK_END);
    certLen = ftell (fCert);
    fseek (fCert, 0, SEEK_SET);
    certBuff = (char * )malloc (certLen);
    if (!certBuff) 
        RslError("訡  %i", certLen);
    fread (certBuff, 1, certLen, fCert);
    fclose(fCert);

    BIO *bio = NULL;

    bio = BIO_new_mem_buf((void*)certBuff, certLen);
    X509 *x509 = PEM_read_bio_X509(bio, NULL, 0, NULL);

    if (x509 == NULL)
        RslError("訡 ⥭ 䨪  䠩  %s",sCertPath);

    free(certBuff);
    BIO_free(bio);

    return x509;
}

// 祭 䨪, ⥪ 䨪 ।  ࠬ
X509 * rsGetX509mem(int iParam) {
    VALUE *vCertText;
    char *sCertPath;

    if (GetParm (iParam, &vCertText) && vCertText->v_type  == V_STRING) { 
        }
    else
        RslError("ࠬ %i   ப c  䠩 䨪",(iParam+1));

    char * certBuff = NULL;
    long certLen = strlen(vCertText->value.string) + sizeof(char);

    BIO *bio = NULL;

    bio = BIO_new_mem_buf((void*)vCertText->value.string, certLen);
    X509 *x509 = PEM_read_bio_X509(bio, NULL, 0, NULL);

    if (x509 == NULL)
        RslError("訡 ⥭ 䨪");

    free(certBuff);
    BIO_free(bio);

    return x509;
}

/*
void pub_key(X509_PUBKEY *xpkey) {
    ASN1_OBJECT *xpoid;
    X509_PUBKEY_get0_param(&xpoid, NULL, NULL, NULL, xpkey);

    exportX509kv_ASN1Obj("PublicKey.Alg", xpoid, vArrayKeys, vArrayValues, &indexElement) {
}
*/

static void opselX509cert(void) {

    VALUE *vArrayKeys;
    if (GetParm (1,&vArrayKeys) && vArrayKeys->v_type  == V_GENOBJ && RslIsTArray((TGenObject *)vArrayKeys->value.obj)) { 
    } 
    else
        RslError("ன ࠬ 㭪樨 opselX509   TArray");

    VALUE *vArrayValues;
    if (GetParm (2,&vArrayValues) && vArrayValues->v_type  == V_GENOBJ && RslIsTArray((TGenObject *)vArrayValues->value.obj)) { 
    }
    else
        RslError("⨩ ࠬ 㭪樨 opselX509   TArray");

    int indexElement=0; // 稪 ⮢  ᨢ vArrayKeys  vArrayValues

    X509 *x509 = rsGetX509(0);

    if (x509 == NULL)
        RslError("訡 ⥭ 䨪");


    //char * sSubj   = X509_NAME_oneline(X509_get_subject_name(x509), NULL, 0);
    //char * sIssuer = X509_NAME_oneline(X509_get_issuer_name(x509), NULL, 0);
    
    exportX509Name(X509_get_subject_name(x509), "subject.", vArrayKeys, vArrayValues, &indexElement);
    exportX509Name(X509_get_issuer_name(x509),  "issuer.",  vArrayKeys, vArrayValues, &indexElement);

    exportX509date(X509_get_notBefore(x509), "not_before", vArrayKeys, vArrayValues, &indexElement);
    exportX509date(X509_get_notAfter(x509),  "not_after",  vArrayKeys, vArrayValues, &indexElement);
    

    /*
    { // serial number
        ASN1_INTEGER * psn=X509_get_serialNumber(x509);     

        if (psn) {

            BIGNUM bn;
            ASN1_INTEGER_to_BN(psn, &bn); //  
            char * sbn = BN_bn2hex(&bn);
            exportX509kv("SerialNumber", sbn, vArrayKeys, vArrayValues, &indexElement); 
            OPENSSL_free(sbn);
        } else
            exportX509kv("SerialNumber", "ERROR", vArrayKeys, vArrayValues, &indexElement); 
    } 
    */

    { // serial number
        ASN1_INTEGER * psn=X509_get_serialNumber(x509);     

        if (psn) {

            BIGNUM * pbn = ASN1_INTEGER_to_BN(psn, NULL); //  
            char * sbn = BN_bn2hex(pbn);
            exportX509kv("SerialNumber", sbn, vArrayKeys, vArrayValues, &indexElement); 
            OPENSSL_free(sbn);
        } else
            exportX509kv("SerialNumber", "ERROR", vArrayKeys, vArrayValues, &indexElement); 
    } 


    fingerprint(x509, vArrayKeys, vArrayValues, &indexElement);



    /*
    int ext_count=X509_get_ext_count(x509);
    print("ext count: %i\n", ext_count);
    for (int i = 0; i < ext_count; i++) {
        X509_EXTENSION * ext =X509_get_ext(x509, i);
        ASN1_OBJECT *o =X509_EXTENSION_get_object(ext);

        char namebuf[X509_NAME_MAXLEN];
        int buflen = OBJ_obj2txt(namebuf, sizeof(namebuf), o, 0);

        ASN1_OCTET_STRING * os = X509_EXTENSION_get_data(ext);

        void * dp =X509V3_EXT_d2i(ext);
        }
    */
    EVP_PKEY *pkey;
    
    //if ((pkey = X509_get_pubkey(x509)) == NULL)
    //    exportX509kv("PublicKey1","NO",vArrayKeys, vArrayValues, &indexElement);
    //else
    //    exportX509kv("PublicKey1","YES",vArrayKeys, vArrayValues, &indexElement);
    //struct X509_pubkey_st:
    //X509_ALGOR *algor;
    //ASN1_BIT_STRING *public_key;
    //EVP_PKEY *pkey;


    X509_PUBKEY *pubkey;
    if ((pubkey = X509_get_X509_PUBKEY(x509)) == NULL)
    //if ((pkey = X509_get_pubkey(x509)) == NULL)
        exportX509kv("PublicKey","NO",vArrayKeys, vArrayValues, &indexElement);
    else {

        do {
        
            size_t pubkey_size = i2d_X509_PUBKEY(pubkey, NULL); // http://groups.google.com/group/mailing.openssl.users/browse_thread/thread/d61858dae102c6c7 
            if(!pubkey_size) {
                exportX509kv("PublicKey","ERR1",vArrayKeys, vArrayValues, &indexElement);
                break; 
            }

            unsigned char * pubkey_buff;
            if ((pubkey_buff = (unsigned char * ) OPENSSL_malloc(pubkey_size))==NULL) {
                exportX509kv("PublicKey","ERR2",vArrayKeys, vArrayValues, &indexElement);
                break;
            }

            if (pubkey_size != i2d_X509_PUBKEY(pubkey, &pubkey_buff)) {
                exportX509kv("PublicKey","ERR3",vArrayKeys, vArrayValues, &indexElement);
                break;
            }


            {
                ASN1_OBJECT *p_alg_oid;
                ASN1_OCTET_STRING *octet = NULL;
                const unsigned char *pubkey_buf = NULL;
                int pubkey_buf_len;

                X509_PUBKEY_get0_param(&p_alg_oid, &pubkey_buf, &pubkey_buf_len, NULL, pubkey);
                exportX509kv_ASN1Obj("PublicKey.Alg", p_alg_oid, vArrayKeys, vArrayValues, &indexElement);

                exportX509kv_int("PublicKey.Size", pubkey_buf_len, vArrayKeys, vArrayValues, &indexElement);
                exportX509kv_bin("PublicKey.ASN1", pubkey_buf_len, pubkey_buf, vArrayKeys, vArrayValues, &indexElement);

                octet = d2i_ASN1_OCTET_STRING(NULL, &pubkey_buf, pubkey_buf_len);
                exportX509kv_bin("PublicKey.Body", octet->length, octet->data, vArrayKeys, vArrayValues, &indexElement);

            }
        } while(0);

    }

    {

        if (X509_check_purpose(x509, X509_PURPOSE_SSL_CLIENT, 0))     exportX509kv("Purpose", "SSL_CLIENT",      vArrayKeys, vArrayValues, &indexElement); 
        if (X509_check_purpose(x509, X509_PURPOSE_SSL_SERVER, 0))     exportX509kv("Purpose", "SSL_SERVER",      vArrayKeys, vArrayValues, &indexElement); 
        if (X509_check_purpose(x509, X509_PURPOSE_NS_SSL_SERVER, 0))  exportX509kv("Purpose", "NS_SSL_SERVER",   vArrayKeys, vArrayValues, &indexElement); 
        if (X509_check_purpose(x509, X509_PURPOSE_SMIME_SIGN, 0))     exportX509kv("Purpose", "SMIME_SIGN",      vArrayKeys, vArrayValues, &indexElement); 
        if (X509_check_purpose(x509, X509_PURPOSE_SMIME_ENCRYPT, 0))  exportX509kv("Purpose", "SMIME_ENCRYPT",   vArrayKeys, vArrayValues, &indexElement); 
        if (X509_check_purpose(x509, X509_PURPOSE_CRL_SIGN, 0))       exportX509kv("Purpose", "CRL_SIGN",        vArrayKeys, vArrayValues, &indexElement); 
        if (X509_check_purpose(x509, X509_PURPOSE_ANY, 0))            exportX509kv("Purpose", "ANY",             vArrayKeys, vArrayValues, &indexElement); 
        if (X509_check_purpose(x509, X509_PURPOSE_OCSP_HELPER, 0))    exportX509kv("Purpose", "OCSP_HELPER",     vArrayKeys, vArrayValues, &indexElement); 
        if (X509_check_purpose(x509, X509_PURPOSE_TIMESTAMP_SIGN, 0)) exportX509kv("Purpose", "TIMESTAMP_SIGN",  vArrayKeys, vArrayValues, &indexElement); 

    }

    X509_free(x509);
    //free(certBuff);
    //BIO_free(bio);
    ReturnVal(V_INTEGER,&indexElement);

}

//static void opselX509cert(void)    {opselX509certCommon(0);}
//static void opselX509certStr(void) {opselX509certCommon(1);}


// 0 - DigestName ப  ⬮ 
// 1 - FilePath   䠩  묨
// 2 -   䨪
// 3 -   
// 4 - ⨯  (0-(def)bin, 1-x16, 2-b64)
static void opselCheckFile(void) {
    const char * sGostPrefix="md_gost";
    EVP_MD_CTX *mdctx;
    const EVP_MD *md;
    unsigned char md_value[EVP_MAX_MD_SIZE];
    long lErrLine;

    VALUE *vFilePath; // vFilePath->value.string
    VALUE *vDigestName;
    char * sSignPath; 

    char* sDigestName;

    if (!GetParm (1,&vFilePath) && vFilePath->v_type  == V_STRING)
        RslError("ࠬ 2   ப");
    char * sPath=(char *)malloc(sizeof(char)*strlen(vFilePath->value.string)+sizeof(char));
    OemToCharBuff(vFilePath->value.string, sPath, strlen(vFilePath->value.string));
    sPath[strlen(vFilePath->value.string)]='\0';

    if (GetParm (0,&vDigestName) && vDigestName->v_type  == V_STRING) { 
        sDigestName=vDigestName->value.string;
    }
    else {
        sDigestName="md5";
    }
    
    X509 *x509 = rsGetX509(2);
    if (x509 == NULL)
        RslError("訡 祭 䨪"); //   

    size_t result;

    FILE * pFile;
    if((pFile = fopen(sPath, "rb"))==NULL)
        RslError("訡  䠩 %s",sPath);

    fseek(pFile, 0, SEEK_END);
    long lFileSize = ftell(pFile);
    rewind (pFile);

    long lBuffSize = lFileSize;
    if (lBuffSize>FILE_MAX_CHUNK_SIZE)
        lBuffSize=FILE_MAX_CHUNK_SIZE;

    char * buffer;
    if((buffer = (char*) malloc (sizeof(char)*lBuffSize))==NULL)
        RslError("訡  %l",lBuffSize);

    //---
    if((sSignPath=rsGetFilePathParam(3))==NULL)
        RslError("訡 祭  "); 

    FILE * pSignFile;
    if((pSignFile= fopen(sSignPath, "rb"))==NULL)
        RslError("訡  䠩 %s",sSignPath);

    fseek(pSignFile, 0, SEEK_END);
    long lSignFileSize = ftell(pSignFile);
    rewind (pSignFile);

    unsigned char * bufferSign;
    if((bufferSign = (unsigned char*) malloc (sizeof(unsigned char)*lSignFileSize))==NULL)
        RslError("訡  %l",lSignFileSize);

    if (!fread(bufferSign, sizeof(unsigned char), lSignFileSize, pSignFile)) 
        RslError("訡 ⥭ 䠩 %s",sSignPath);

    fclose(pSignFile);

    //---
    ENGINE *eGost = NULL;

    if(!OPENSSL_init_crypto(OPENSSL_INIT_ENGINE_ALL_BUILTIN | OPENSSL_INIT_ENGINE_DYNAMIC, NULL))
        RslError("訡  樠樨 ਯ⮡⥪");
    
    if (strlen(sDigestName)>=strlen(sGostPrefix) && 0==memcmp(sGostPrefix,sDigestName,strlen(sGostPrefix)))
        if((eGost = ENGINE_by_id("gost"))==NULL) 
            RslError(" 㦥    openssl 室  -㭪樨 %s",sDigestName);
        
    md = EVP_get_digestbyname(sDigestName);
    if (!md)
        RslError("-㭪 %s  ", sDigestName);

    mdctx = EVP_MD_CTX_new();

    EVP_PKEY_CTX *pctx = NULL;
    EVP_PKEY *pkey = NULL;

    //---

    X509_PUBKEY *pubkey;
    if ((pubkey = X509_get_X509_PUBKEY(x509)) == NULL)
        RslError("訡 祭 㡫筮   䨪");

    size_t pubkey_size = i2d_X509_PUBKEY(pubkey, NULL); // http://groups.google.com/group/mailing.openssl.users/browse_thread/thread/d61858dae102c6c7 
    if (!pubkey_size) {
        RslError("PublicKey ERR1");
    }

    unsigned char * pubkey_buff;
    if ((pubkey_buff = (unsigned char * ) OPENSSL_malloc(pubkey_size))==NULL) {
        RslError("PublicKey ERR2");
    }

    if (pubkey_size != i2d_X509_PUBKEY(pubkey, &pubkey_buff)) {
        RslError("PublicKey ERR3");
    }


    if ((pkey = X509_PUBKEY_get(pubkey)) == NULL)
        RslError("訡 祭 㡫筮   䨪");
    //---

    if(!EVP_DigestVerifyInit(mdctx, &pctx, md, eGost, pkey))
        RslError("訡 樠樨   ஢ન ");

    while (size_t chunkSize=fread(buffer, sizeof(char), lBuffSize, pFile)) {
        //EVP_DigestUpdate(mdctx, buffer, chunkSize);
        EVP_DigestVerifyUpdate(mdctx, buffer, chunkSize);
    }

    int ret;
    unsigned int md_len;
    EVP_DigestFinal_ex(mdctx, md_value, &md_len);
    if(1==EVP_DigestVerifyFinal(mdctx, bufferSign, lSignFileSize)) {
        ret=1;
    }
    else {
        ret=0;
    }

    EVP_MD_CTX_free(mdctx);
    fclose (pFile);
    free (buffer);
    free(bufferSign);

    ReturnVal(V_BOOL,&ret);
}



EXP32 void DLMAPI EXP AddModuleObjects (void)
{
    AddStdProc (V_STRING, "opselHash",        opselHash,        0);
    AddStdProc (V_STRING, "opselHashFile",    opselHashFile,    0);
    AddStdProc (V_INTEGER,"opselX509cert",    opselX509cert,    0);
//    AddStdProc (V_INTEGER,"opselX509certStr", opselX509certStr, 0);
    AddStdProc (V_INTEGER,"opselCheckFile",   opselCheckFile,   0);
}                     


EXP32 void DLMAPI EXP InitExec (void)
{
}

EXP32 void DLMAPI EXP DoneExec (void)
{
}
